//Implementation of Omega-R.AdaBoost vs AdaBoost, NIPS 2016 code

import java.util.*;
import java.io.*;

class Algorithm implements Debuggable{
    
    Vector all_algorithms;
    Domain myDomain;
    String classifierToKeep;

    Algorithm(Domain dom, String kkk){
	all_algorithms = new Vector();
	myDomain = dom;
	classifierToKeep = kkk;

	Dataset.warning("Classifier kept in sequence for all algorithms: " + kkk + "\n");
    }

    public static String ADABOOST = "@AdaBoost";
    public static String RADOBOOST = "@RadoBoost";
    public static String EXAMPLES_TRAIN_SIZE = "@TRAIN_SIZE", EXAMPLES_MAX = "@MAX", EXAMPLES_EXACTLY = "@EXACTLY";
    // TRAIN_SIZE = for examples, we take all training data; for rados, we contider a number of rados equal to training examples size
    // MAX = we take min {training fold / 2, integer indicated in @MAX_RADO_SIZE}

    public static String [] REGULARIZATION_OMEGA = {"@N_A", "@NO", "@L22", "@L1", "@LINF", "@LSLOPE"};

    public static int INDEX_REGULARIZATION(String s){
	int i = 0;
	do{
	    if (REGULARIZATION_OMEGA[i].equals(s))
		return i;
	    i++;
	}while (i<REGULARIZATION_OMEGA.length);
	Dataset.perror("Alrgorithm.class :: No regularization method for parameter " + s);
	return -1;
    }

    public static void CHECK_REGULARIZATION(LinearBoost lb){
	if ( (!lb.isRadoType) && (lb.isRegularized) && ( (lb.indexRegularized == 2) || (lb.indexRegularized == 4) || (lb.indexRegularized == 5) ) )
	    Dataset.perror("Algorithm.class :: Regularization " + REGULARIZATION_OMEGA[lb.indexRegularized] + " not available");
    }

    public void addAlgorithm(Vector all_params){
	int kkki = Dataset.INDEX_KEY_CLASSIFIER_KEPT(classifierToKeep);

	String d = (String) all_params.elementAt(0), sRows, gOmega = null;
	int nIter, nMu, nWeight, nMethod, nExact = -1, delta, index = 1;
	double gRegularized = -1.0;
	boolean bRegularized = false;

	nIter = Integer.parseInt((String) all_params.elementAt(index));
	index++;

	bRegularized = Boolean.parseBoolean((String) all_params.elementAt(index));
	index++;
	gOmega = (String) all_params.elementAt(index);
	index++;
	gRegularized = Double.parseDouble((String) all_params.elementAt(index));
	index++;

	nMu = Integer.parseInt((String) all_params.elementAt(index));
	index++;
	nWeight = Integer.parseInt((String) all_params.elementAt(index));
	index++;

	sRows = (String) all_params.elementAt(index);
	index++;

	if ( !(sRows.equals(Algorithm.EXAMPLES_TRAIN_SIZE)) && !(sRows.equals(Algorithm.EXAMPLES_MAX)) && !(sRows.equals(Algorithm.EXAMPLES_EXACTLY)) )
	    Dataset.perror("Algorithm :: Keyword " + sRows + " not authorized for number of examples/rados in resource file");
	
	if (sRows.equals(Algorithm.EXAMPLES_EXACTLY)){
	    nExact = Integer.parseInt((String) all_params.elementAt(index));
	    //index++;

	    delta = 1;
	}else
	    delta = 0;

	if (d.equals(Algorithm.ADABOOST)){
	    all_algorithms.addElement(new LinearBoost(myDomain, d, nIter, bRegularized, gOmega, gRegularized, nMu, nWeight, sRows, nExact, kkki));
	}else if (d.equals(Algorithm.RADOBOOST)){
	    if ( ( (delta == 0) && (all_params.size() == 9) ) || ( (delta == 1) && (all_params.size() == 10) ) )
		all_algorithms.addElement(new LinearBoost(myDomain, d, nIter, bRegularized, gOmega, gRegularized, nMu, nWeight, sRows, nExact, (String) all_params.elementAt(index+delta), kkki));
	    else
		all_algorithms.addElement(new LinearBoost(myDomain, d, nIter, bRegularized, gOmega, gRegularized, nMu, nWeight, sRows, nExact, (String) all_params.elementAt(index+delta), Double.parseDouble((String) all_params.elementAt(index+1+delta)), kkki));
	}else
	    Dataset.perror("Algorithm :: Keyword " + d + " not authorized for algorithm in resource file");
    }

    public Vector go(){
	int i;
	Vector vcur, vret = new Vector();
	for (i=0;i<all_algorithms.size();i++){
	    vcur = ((LinearBoost) all_algorithms.elementAt(i)).boost();
	    vret.addElement(vcur);
	}
	return vret;
    }

    public static boolean contains(int [] ii, int d){
	int i;
	for (i=0;i<ii.length;i++)
	    if (ii[i] == d)
		return true;
	return false;
    }

    public void save(Vector v){
	int [] indexLatexSaveRandom = {3, 4};
	int [] indexLatexSaveClassWise = {3, 29};

	Calendar cal = Calendar.getInstance();

	String now = cal.get(Calendar.DAY_OF_MONTH) + "th__" + cal.get(Calendar.HOUR_OF_DAY) + "h_" + cal.get(Calendar.MINUTE) + "m_" + cal.get(Calendar.SECOND) + "s";
	String nameSave = myDomain.myDS.pathSave + "results_" + now + ".txt";
	int i, j, k, ir = -1, irt = -1;
	Vector vi, vj;

	FileWriter f = null;

	int smax = 7;
	int index_err_test = 4;
	int index_supporteps = 6;

	String [] names = {"Tmax            ", "BestIter          ", "Err_Emp_Init (*100)", "Min_Err_Emp (*100)", "Err_Test (*100)    ", "L2_Norm          ", "Support (*100) (%)   "};

	double [] entries;
	double [] entries2;
	double [] as;

	double tp;

	boolean rec = false, brt = false, alo = false;

	int [][] index_best_radoboost = new int [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<index_best_radoboost.length;i++){
	    index_best_radoboost[i] = new int [Domain.RADO_GENERATE.length];
	    for (j=0;j<index_best_radoboost[i].length;j++)
		index_best_radoboost[i][j] = -1;
	}

	double [][] test_err_avg_best_radoboost = new double [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<test_err_avg_best_radoboost.length;i++){
	    test_err_avg_best_radoboost[i] = new double [Domain.RADO_GENERATE.length];
	    for (j=0;j<test_err_avg_best_radoboost[i].length;j++)
		test_err_avg_best_radoboost[i][j] = -1.0;
	}

	double [][] test_err_stdev_best_radoboost = new double [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<test_err_stdev_best_radoboost.length;i++){
	    test_err_stdev_best_radoboost[i] = new double [Domain.RADO_GENERATE.length];
	    for (j=0;j<test_err_stdev_best_radoboost[i].length;j++)
		test_err_stdev_best_radoboost[i][j] = -1.0;
	}

	boolean [][] met_best_radoboost = new boolean [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<met_best_radoboost.length;i++){
	    met_best_radoboost[i] = new boolean [Domain.RADO_GENERATE.length];
	    for (j=0;j<met_best_radoboost[i].length;j++)
		met_best_radoboost[i][j] = false;
	}
	    
	double [][] support_avg_best_radoboost = new double [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<support_avg_best_radoboost.length;i++){
	    support_avg_best_radoboost[i] = new double [Domain.RADO_GENERATE.length];
	    for (j=0;j<support_avg_best_radoboost[i].length;j++)
		support_avg_best_radoboost[i][j] = -1.0;
	}

	double [][] support_stdev_best_radoboost = new double [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<support_stdev_best_radoboost.length;i++){
	    support_stdev_best_radoboost[i] = new double [Domain.RADO_GENERATE.length];
	    for (j=0;j<support_stdev_best_radoboost[i].length;j++)
		support_stdev_best_radoboost[i][j] = -1.0;
	}

	double [][] test_err_avg_worst_radoboost = new double [Algorithm.REGULARIZATION_OMEGA.length][];
	for (i=0;i<test_err_avg_worst_radoboost.length;i++){
	    test_err_avg_worst_radoboost[i] = new double [Domain.RADO_GENERATE.length];
	    for (j=0;j<test_err_avg_worst_radoboost[i].length;j++)
		test_err_avg_worst_radoboost[i][j] = -1.0;
	}

	int out3 = 0, out10 = 0, out100 = 0;
	for (i=0;i<myDomain.myDS.outlier_features_3s.length;i++){
	    if (myDomain.myDS.outlier_features_3s[i])
		out3++;
	    if (myDomain.myDS.outlier_features_10s[i])
		out10++;
	    if (myDomain.myDS.outlier_features_100s[i])
		out100++;
	}

	try{
	    f = new FileWriter(nameSave);
	    f.write("%Domain " + myDomain.myDS.domainName + " with classes centered wrt " + Dataset.FIT_CLASS_MODALITIES[Dataset.DEFAULT_INDEX_FIT_CLASS] + "\n");
	    f.write("%Proportion of sign(+/-) in classes : \t" + DF.format(myDomain.myDS.getProportionExamplesSign(true)) + "/" + DF.format(myDomain.myDS.getProportionExamplesSign(false)) + "\n");
	    f.write("%Domain #examples : \t\t\t" + myDomain.myDS.number_examples_total + "\n");
	    f.write("%Domain #features : \t\t\t" + myDomain.myDS.number_initial_features + "\n");
	    f.write("%features (3*sigma) outliers   : \t" + out3 + "/" + myDomain.myDS.outlier_features_3s.length + "\n");
	    f.write("%features (10*sigma) outliers  : \t" + out10 + "/" + myDomain.myDS.outlier_features_3s.length + "\n");
	    f.write("%features (100*sigma) outliers : \t" + out100 + "/" + myDomain.myDS.outlier_features_3s.length + "\n\n");

	    f.write("%classifier kept in sequence : \t\t" + classifierToKeep + "\n\n");

	    f.write("% Single algorithms statistics ::\n");
	    for (i=0;i<all_algorithms.size();i++){
		f.write("%Algo_" + i + " = " + ((LinearBoost) all_algorithms.elementAt(i)).fullName() + "\n");
		vi = (Vector) v.elementAt(i);
		if (vi.size() != NUMBER_STRATIFIED_CV)
			Dataset.perror("not the right statistics in vector " + i);

		rec = false;
		for (j=0;j<smax;j++){
		    entries = new double[NUMBER_STRATIFIED_CV];
		    for (k=0;k<NUMBER_STRATIFIED_CV;k++)
			entries[k] = ((Double) ((Vector) vi.elementAt(k)).elementAt(j)).doubleValue();
		    as = new double[2];
		    Statistics.avestd(entries, as);
		    f.write(names[j] + "\t" + DF.format(as[0]) + "\t\\pm\t" + DF.format(as[1]) + "\n");

		    if (j==index_err_test){
			ir = ((LinearBoost) all_algorithms.elementAt(i)).indexRegularized;

			if (((LinearBoost) all_algorithms.elementAt(i)).methodRados != null)
			    irt = Domain.INDEX_GENERATE(((LinearBoost) all_algorithms.elementAt(i)).methodRados);
			brt = ((LinearBoost) all_algorithms.elementAt(i)).isRadoType;

			if ( (brt) && (((LinearBoost) all_algorithms.elementAt(i)).methodRados == null) )
			    Dataset.perror("Algorithm.class :: Algorithm " + i + " is rado type but has rado method null");

			if ( (brt) && (ir > 1) ){
			    // otherwise, no regularisation or N/A or no rado
			    if (!met_best_radoboost[ir][irt]){
				index_best_radoboost[ir][irt] = i;
				test_err_avg_best_radoboost[ir][irt] = as[0];
				test_err_avg_worst_radoboost[ir][irt] = as[0];
				test_err_stdev_best_radoboost[ir][irt] = as[1];
				met_best_radoboost[ir][irt] = true;
				rec = true;
			    }else if ( (met_best_radoboost[ir][irt]) && (test_err_avg_best_radoboost[ir][irt] > as[0]) ){
				index_best_radoboost[ir][irt] = i;
				test_err_avg_best_radoboost[ir][irt] = as[0];
				test_err_stdev_best_radoboost[ir][irt] = as[1];
				rec = true;
			    }else if ( (met_best_radoboost[ir][irt]) && (test_err_avg_worst_radoboost[ir][irt] < as[0]) ){
				test_err_avg_worst_radoboost[ir][irt] = as[0];
			    }
			}
		    }
		    else if (j==index_supporteps){
			if (rec){
			    ir = ((LinearBoost) all_algorithms.elementAt(i)).indexRegularized;
			    support_avg_best_radoboost[ir][irt] = as[0];
			    support_stdev_best_radoboost[ir][irt] = as[1];
			}
		    }
		}
		f.write("\n");
	    }

	    f.write("\n% Test errs paired t-tests ::\n");
	    for (i=0;i<all_algorithms.size()-1;i++)
		for (j=i+1;j<all_algorithms.size();j++){
		    f.write("%Algo_" + i + " = " + ((LinearBoost) all_algorithms.elementAt(i)).fullName() + "\n");
		    f.write("%Algo_" + j + " = " + ((LinearBoost) all_algorithms.elementAt(j)).fullName() + "\n");

		    vi = (Vector) v.elementAt(i);
		    vj = (Vector) v.elementAt(j);
		    
		    entries = new double[NUMBER_STRATIFIED_CV];
		    entries2 = new double[NUMBER_STRATIFIED_CV];

		    for (k=0;k<NUMBER_STRATIFIED_CV;k++){
			entries[k] = ((Double) ((Vector) vi.elementAt(k)).elementAt(index_err_test)).doubleValue(); //ERR TEST
			entries2[k] = ((Double) ((Vector) vj.elementAt(k)).elementAt(index_err_test)).doubleValue(); //ERR TEST
		    }

		    tp = Statistics.tptest(entries, entries2);
		    f.write("P-val = " + tp + "\n\n");
		}	    

	    for (j=0;j<Domain.RADO_GENERATE.length;j++){
		alo = false;
		for (i=2;i<Algorithm.REGULARIZATION_OMEGA.length;i++)
		    if (met_best_radoboost[i][j])
			alo = true;
		if (alo){
		    f.write("\n\nBest regularization parameters and algorithms for RADO_TYPE = " + Domain.RADO_GENERATE[j] + " : \n");
		    for (i=2;i<Algorithm.REGULARIZATION_OMEGA.length;i++){
			if (met_best_radoboost[i][j]){
			    f.write("Reg.[" + Algorithm.REGULARIZATION_OMEGA[i] + "]:\t" + DF0.format(test_err_avg_best_radoboost[i][j]) + "\t$\\pm$\t" + DF0.format(test_err_stdev_best_radoboost[i][j]) + " (support =\t" + DF0.format(support_avg_best_radoboost[i][j]) + "\t$\\pm$\t" + DF0.format(support_stdev_best_radoboost[i][j]) + "\t) --- Delta = " + DF0.format((test_err_avg_worst_radoboost[i][j]-test_err_avg_best_radoboost[i][j])) + ", for best instance: " + ((LinearBoost) all_algorithms.elementAt(index_best_radoboost[i][j])).fullName() + "\n");
			}
		    }
		}
	    }

	    f.write("\n\n");

	    f.close();
	}catch(IOException e){
	    Dataset.perror("LinearBoost.class :: Saving results error in file " + nameSave);
	}
    }
}
