//Implementation of Omega-R.AdaBoost vs AdaBoost, NIPS 2016 code

import java.io.*;
import java.util.*;


class Experiments implements Debuggable{
    
    public static boolean LIMIT_RADO_NUMBER = false;
    // if true, applies the max{train/2, 1000} size limit for rados, as in Nock, Patrini, Friedman ICML 2015 paper "Rademacher Observations, Private Data and Boosting"

    public static String HISTORY_DATE[] = {"02/09/2015", "29/10/2015", "08/04/2016", "28/09/2016"};
    public static String HISTORY_VERSION[] = {"V0.1","V0.2", "V0.3", "V1.0"};
    public static String HISTORY_WHATSNEW[] = {"Includes privacy related content + (Gamma)L22 regularisation",
					       "Added L1-AdaBoost, plus L1, Linfinity, Slope regularisations for Omega.R.AdaBoost + class-wise rados",
					       "Brushed a bit the code",
					       "Excluded privacy related content to match exactly the NIPS paper's content"};

    public static boolean FAST_PROCESSING = true;
    // skips some computations for faster processing times, but no approximations

    public static String KEY_HELP = "--help", KEY_RESOURCE = "-R";
    
    Algorithm myAlgos;
    Domain myDomain;

    public static String help(){
	String ret = "";
	ret += KEY_HELP + " : example command line\n\n";
	ret += KEY_RESOURCE + " :: name of resource file to parse algorithms\n";

	return ret;
    }

    public static void main(String [] arg){
	int i;
	String kR = "";
	for (i=0;i<arg.length;i++){
	    if (arg[i].equals(KEY_HELP)){
		Dataset.perror(help());
	    }

	    if (arg[i].equals(KEY_RESOURCE))
		kR = arg[i+1];
	}
	
	if (kR.equals(new String("")))
	    Dataset.perror("No resource file name found in command line");


	System.out.println("\n** Regularized RadoBoost (Version NIPS 2016) \n** Consult the README.txt file\n-- Contact: richard.nock@data61.csiro.au\n");

	System.out.print("\n** Parsing resource file: " + kR + " ...\n");

	Experiments ee = new Experiments();
	ee.go(kR);
    }

    public void go(String rs){
	Vector v;
	parse(rs);
	v = myAlgos.go();
	myAlgos.save(v);

	myDomain.myMemoryMonitor.stop();
    }
    
    public void parse(String rs){
	FileReader e;
	BufferedReader br;
	StringTokenizer t;
	String dum, n, nameD = "", nameP = "", kkk = "";
	Vector v;
	int maxr = -1;
	double q_slope = -1.0;

	myDomain = null;

	// Domain
	try{
	    e = new FileReader(rs);
	    br = new BufferedReader(e);
	    
	    while ( (dum=br.readLine()) != null){
		if ( (dum.length() == 1) || ( (dum.length()>1) && (!dum.substring(0,Dataset.KEY_COMMENT.length()).equals(Dataset.KEY_COMMENT)) ) ){
		    t = new StringTokenizer(dum,Dataset.KEY_SEPARATION_STRING[Dataset.SEPARATION_INDEX]); 
		    n = t.nextToken();
		    if (n.equals(Dataset.KEY_DIRECTORY))
			nameD = t.nextToken();
		    else if (n.equals(Dataset.KEY_PREFIX))
			nameP = t.nextToken();
		    else if (n.equals(Dataset.MAX_RADO))
			maxr = Integer.parseInt(t.nextToken());
		    else if (n.equals(Dataset.Q_SLOPE))
			q_slope = Double.parseDouble(t.nextToken());
		    else if (n.equals(Dataset.KEY_CLASSIFIER_KEPT))
			kkk = t.nextToken();
		}
	    }
	    e.close();
	}catch(IOException eee){
	    System.out.println("Problem loading ." + rs + " resource file --- Check the access to file");
	    System.exit(0);
	}

	if (nameD.equals(new String("")))
	    Dataset.perror("No domain in resource file");
	if (nameP.equals(new String("")))
	    Dataset.perror("No prefix in resource file");
	if (maxr == -1)
	    Dataset.perror("No value specified for max rado size in resource file");

	System.out.println("\n/////////////////////////////////////////////\n/// Domain " + nameP + " in directory " + nameD + "\n/////////////////////////////////////////////\n");

	myDomain = new Domain(nameD, nameP, maxr, q_slope);
	myAlgos = new Algorithm(myDomain, kkk);

	// Algos
	try{
	    e = new FileReader(rs);
	    br = new BufferedReader(e);
	    
	    while ( (dum=br.readLine()) != null){
		if ( (dum.length() == 1) || ( (dum.length()>1) && (!dum.substring(0,Dataset.KEY_COMMENT.length()).equals(Dataset.KEY_COMMENT)) ) ){
		    t = new StringTokenizer(dum,Dataset.KEY_SEPARATION_STRING[Dataset.SEPARATION_INDEX]); 
		    n = t.nextToken();
		    if (n.equals(Dataset.KEY_ALGORITHM)){
			v = new Vector();
			while(t.hasMoreTokens())
			    v.addElement(new String(t.nextToken()));
			myAlgos.addAlgorithm(v);
		    }
		}
	    }
	    e.close();
	}catch(IOException eee){
	    System.out.println("Problem loading ." + rs + " resource file --- Check the access to file");
	    System.exit(0);
	}
    }

}
