//Implementation of Omega-R.AdaBoost vs AdaBoost, NIPS 2016 code

import java.util.*;
import java.io.*;

public class LinearBoost implements Debuggable{
    public static int MAX_BOOSTING_ITERS = 100;

    public static String [] METHOD_WEIGHTS_STRING = {"AdaBoost.R", "Real Adaboost (S.S.)"};
    // COMMENT:
    // AdaBoost S.S. is Schapire-Singer's AdaBoost
    // AdaBoost.R is Nock-Nielsen's "A Real Generalization of Discrete AdaBoost", in Artificial Intelligence (2007)

    public static  double SCALE_MU_THRESHOLD = 0.1; 
    public static String [] METHOD_MU_STRING = {"Edge used as is", "Edge scaled so its |.| is >= " + SCALE_MU_THRESHOLD + " "};

    public static double MAX_ABS_MU = 0.8;

    public static double DEFAULT_GAMMA_REGULARISATION = 0.5;

    Domain myDomain;
    
    int split_number;

    double [] theta;
    double [] weights;
    double [] star;

    Vector all_classifiers; 
    // first dim = fold number
    // second dim = iteration
    // third dim = classifier

    int methodWeights, nIterMax, methodMu, nRows, nExact, nClassifierKept; //nExact is the "ideal" value if specified of the number of examples

    // Regularization parameters
    int indexRegularized;
    boolean isRegularized;
    double gammaRegularized;
    // indexRegularized = index in Algorithm.REGULARIZATION_OMEGA that gives the regularization (for rados ONLY)

    String methodRows, methodRados, name;

    boolean isRadoType;
    // true iff of RadoBoost type

    LinearBoost(Domain dom, String type, int nIter, boolean bRegularized, String gOmega, double gRegularized, int nMu, int nWeight, String sRows, int nEx, int kkki){
	all_classifiers = new Vector();

	myDomain = dom;
	if (type.equals(Algorithm.RADOBOOST))
	    isRadoType = true;
	else if (type.equals(Algorithm.ADABOOST))
	    isRadoType = false;
	else
	    Dataset.perror("Keyword " + type + " not authorized for algorithm in resource file");

	nClassifierKept = kkki;
	name = type;
	nIterMax = nIter;
	methodWeights = nWeight;
	methodMu = nMu;
	methodRows = sRows;
	methodRados = null;
	nRows = -1;
	nExact = nEx;
	weights =null;

	isRegularized = bRegularized;
	if (!isRegularized){
	    gammaRegularized = -1.0;
	    indexRegularized = 0;
	}
	else{
	    gammaRegularized = gRegularized;
	    indexRegularized = Algorithm.INDEX_REGULARIZATION(gOmega);
	    Algorithm.CHECK_REGULARIZATION(this);
	}
    }


    LinearBoost(Domain dom, String type, int nIter, boolean bRegularized, String gOmega, double gRegularized, int nMu, int nWeight, String sRows, int nExact, String sRados, int kkki){
	this(dom, type, nIter, bRegularized, gOmega, gRegularized, nMu, nWeight, sRows, nExact, kkki);
	methodRados = sRados;
    }


    LinearBoost(Domain dom, String type, int nIter, boolean bRegularized, String gOmega, double gRegularized, int nMu, int nWeight, String sRows, int nExact, String sRados, double dd, int kkki){
	this(dom, type, nIter, bRegularized, gOmega, gRegularized, nMu, nWeight, sRows, nExact, sRados, kkki);
    }

    public String regularizationName(char c){
	String ret = "regularisation" + c + "" + Algorithm.REGULARIZATION_OMEGA[indexRegularized];
	if (isRegularized)
	    ret += "" + c + "(gamma_reg" + c + "" + gammaRegularized + ")";
	return ret;
    }

    public String fullName(){
	String ret = name;
	if (methodRados != null)
	    ret += "(" + methodRados + ")";
	ret += " - Iter = " + nIterMax + ", " + regularizationName('=');
	ret +=  ", weights = " + METHOD_WEIGHTS_STRING[methodWeights] + ", mu = " + METHOD_MU_STRING[methodMu] + ", methodRows = " + methodRows + ", nRowsLearning = " + nRows + ", nExact = " + nExact;
	return ret;
    }

    public String fullName_Cryptic(){
	String ret = name;
	if (methodRados != null)
	    ret += "_" + methodRados + "_";
	ret += " - Iter = " + nIterMax + ", " + regularizationName('_');
	ret +=  "weights" + METHOD_WEIGHTS_STRING[methodWeights] + "_mu" + METHOD_MU_STRING[methodMu] + "_methodRows" + methodRows + "_nRowsLearning" + nRows + "_nExact" + nExact;
	return ret;
    }

    public String fullName_Cryptic_short(){
	String ret = name;
	if (methodRados != null)
	    ret += "_" + methodRados + "_";
	ret += " - Iter = " + nIterMax + ", " + regularizationName('_');
	ret +=  "weights" + METHOD_WEIGHTS_STRING[methodWeights] + "_methodRows" + methodRows + "_nRowsLearning" + nRows + "_nExact" + nExact;
	return ret;
    }

    // display methods

    public String theta(){
	String val = "";
	int i;
	val += ("Theta = ");
	for (i=0;i<theta.length;i++)
	    val += (DF.format(theta[i]) + " ");
	return val;
    }

    // Boosting algorithms

    public void initRound(int split){
	theta = new double [myDomain.myDS.number_real_features];
	star = new double [myDomain.myDS.number_real_features];

	System.out.print(" * " + name + "[" + Algorithm.REGULARIZATION_OMEGA[indexRegularized] + "] > F#" + (split+1) + " -- ");
	split_number = split;	

	if (methodRows.equals(Algorithm.EXAMPLES_TRAIN_SIZE))
	    nRows = myDomain.myDS.train_size(split_number);
	else if (methodRows.equals(Algorithm.EXAMPLES_MAX)){
	    if ( (Experiments.LIMIT_RADO_NUMBER) && (myDomain.MAX_RADO_SIZE > myDomain.myDS.train_size(split_number) / 2) )
		nRows = myDomain.myDS.train_size(split_number) / 2;
	    else
		nRows = myDomain.MAX_RADO_SIZE;
	}else if (methodRows.equals(Algorithm.EXAMPLES_EXACTLY)){
	    if (nExact > myDomain.myDS.train_size(split_number))
		nRows = myDomain.myDS.train_size(split_number);
	    else
		nRows = nExact;
	}else
	    Dataset.perror("initRound :: unauthorized value " + methodRows);


	int i, nr, ne = nRows;
	for (i=0;i<theta.length;i++){
	    theta[i] = 0.0;
	}

	if (isRadoType){
	    myDomain.rado_generate(split_number, this);

	    nr = myDomain.rados.size();
	    weights = new double [nr];
	    for (i=0;i<nr;i++)
		weights[i] = 1.0 / ((double) nr);
	    for (i=0;i<theta.length;i++){
		star[i] = myDomain.radostar(this, i);
	    }
	}else{
	    weights = new double [ne];
	    for (i=0;i<ne;i++){
		weights[i] = 1.0 / ((double) ne);
	    }
	    for (i=0;i<theta.length;i++){
		if (nRows == myDomain.myDS.train_size(split_number))
		    star[i] = myDomain.myDS.xstar(this, split_number, i, -1);
		else if (nRows > myDomain.MAX_RADO_SIZE)
		    Dataset.perror("Dataset.class :: please give all training or exactly MAX_RADO_SIZE number of example to train with");
		else
		    star[i] = myDomain.myDS.xstar(this, split_number, i, nRows);
	    }
	}
    }

    public Vector boost(){
	int i;
	Vector cur, res = new Vector();

	for (i=0;i<NUMBER_STRATIFIED_CV;i++){
	    initRound(i);
	    cur = iterate(i);
	    res.addElement(cur);
	}
	System.out.println("");

	if (SAVE_MEMORY)
	    theta = star = weights = null;

	return res;
    }


    public void displayProcess(int i, int nmax){
	if (i%(nmax/10) == 0){
	    System.out.print(".");//System.out.print( ((i/(nmax/5)) * 20) + "%");
	    System.out.print(myDomain.memStringThreshold() + "");
	}
    }

    public Vector iterate(int split){

	Vector ret = new Vector();

	int i, besti = -1, Tmax = 0;
	double err_emp, min_err_emp = -1.0, err_test, get_err_test = -1.0, l2, supporteps, err_emp_init;
	double [] bestTheta = null;
	boolean improvement, keepc;

	Vector dumv, dumv2;
	String binSup;

	err_emp_init = error(true);

	System.out.print("learning[" + nIterMax + "@ITERs");

	for (i=0;i<nIterMax;i++){
	    improvement = one_iteration();
	    keepc = false;
	    
	    if ( (nClassifierKept == 0) && (Experiments.FAST_PROCESSING) ){
		bestTheta = Rado.copyOf(theta); 
		besti = i;
		keepc = true;
	    }else{
		err_emp = error(true);
		if ( (nClassifierKept == 0) || (i==0) || (err_emp <= min_err_emp) ) {
		    err_test = error(false);
		    bestTheta = Rado.copyOf(theta); 
		    min_err_emp = err_emp;
		    get_err_test = err_test;
		    besti = i;
		    keepc = true;
		}
	    }

	    displayProcess(i, nIterMax);

	    Tmax = (i+1);
	    if (!improvement)
		i = nIterMax;
	}
	System.out.print("ok.] ");

	if ( (nClassifierKept == 0) && (Experiments.FAST_PROCESSING) ){
	    min_err_emp = -1.0; // because fast processing, no recording of the best err_emp in sequence
	    get_err_test = error(false);
	}

	l2 = Rado.L2(bestTheta);
	supporteps = Rado.support(bestTheta);
	binSup = Rado.binarize(bestTheta);

	System.out.println("\t(Eerr Terr l2 sup boolsup) = (" + DF0.format( ((100.0) * min_err_emp) ) + " " + DF0.format( ((100.0) * get_err_test) ) + " " + DF8.format(l2) + " " + DF0.format(supporteps) + " " + binSup + ")");

	ret.addElement(new Double(Tmax));
	ret.addElement(new Double((double) besti));
	ret.addElement(new Double(err_emp_init * 100.0));
	ret.addElement(new Double(min_err_emp * 100.0));
	ret.addElement(new Double(get_err_test * 100.0));
	ret.addElement(new Double(l2));
	ret.addElement(new Double(supporteps * 100.0));

	return ret;
    }

    public double rado_regularized_penalty_GammaL22(double mu, int f){
	double pena = 1.0;
	double [] deltav = new double [myDomain.dimOperator];
	double [] nexttheta;
	if (deltav.length != theta.length)
	    Dataset.perror("update delta_t not of the same dimension as classifier theta_t");

	deltav[f] = alphaBoost(muClamp(mu), f);
	nexttheta = Rado.add(deltav, theta);
	double scal = gammaRegularized * (Rado.GammaL22(nexttheta,myDomain.myDS.feature_weights) - Rado.GammaL22(theta,myDomain.myDS.feature_weights));

	return scal;
    }

    public double rado_regularized_penalty_All(double mu, int f, int regularization){
	double pena = 1.0, fact = -1.0;
	double [] deltav = new double [myDomain.dimOperator];
	double [] nexttheta;
	if (deltav.length != theta.length)
	    Dataset.perror("update delta_t not of the same dimension as classifier theta_t");

	deltav[f] = alphaBoost(muClamp(mu), f);
	nexttheta = Rado.add(deltav, theta);

	if (regularization == 3)
	    fact = Rado.L1(nexttheta) - Rado.L1(theta);
	else if (regularization == 4)
	    fact = Rado.LINF(nexttheta) - Rado.LINF(theta);
	else if (regularization == 5)
	    fact = Rado.LSLOPE(myDomain, nexttheta) - Rado.LSLOPE(myDomain, theta);
	else 
	    Dataset.perror("LinearBoost.class :: no regularisation for token " + regularization);

	double scal = gammaRegularized * fact;

	return scal;
    }


    public double rado_regularized_penalty(double mu, int f){
	if (!isRegularized)
	    Dataset.perror("LinearBoost.class :: Method not regularized");
	if (indexRegularized == 2)
	    return rado_regularized_penalty_GammaL22(mu, f);
	else if ( (indexRegularized >= 3) && (indexRegularized <= 5) )
	    return rado_regularized_penalty_All(mu, f, indexRegularized);
	Dataset.perror("LinearBoost.class :: regularization not yet implemented");
	return -1.0;
    }

    public double example_regularized_penalty_L1(double mu, int f){
	double pena = 1.0;
	double [] deltav = new double [myDomain.dimOperator];
	double [] nexttheta;
	if (deltav.length != theta.length)
	    Dataset.perror("update delta_t not of the same dimension as classifier theta_t");

	deltav[f] = alphaBoost(muClamp(mu), f);
	nexttheta = Rado.add(deltav, theta);
	double scal = gammaRegularized * Rado.L1(nexttheta);

	return scal;
    }

    public Vector wfi(){ 
	//implements (Regularized) Weak Feature Index oracle
	int i, iret = -1;
	double cur, mx = 0.0, curmu, optmu = 0.0, pena = 0.0, optpena = 0.0;
	boolean reg = false;
	for (i=0;i<myDomain.dimOperator;i++){
	    curmu = muBoost(i);

	    // Location of regularization for R.AdaBoost
	    if ((isRadoType) && (isRegularized)){
		pena = rado_regularized_penalty(curmu, i);
		reg = true;
	    }

	    if ( (!isRegularized) || (!isRadoType) )
		cur = Math.abs(curmu);
	    else{
		if (!reg)
		    Dataset.perror("Dataset.class :: Regularized method but no regularisation coefficient computed");
		cur = ((curmu*curmu)/2.0) - pena;
	    }

	    if ( (i==0) || (cur > mx) ){
		iret = i;
		mx = cur;
		optmu = curmu;
		if (isRegularized)
		    optpena = pena;
	    }
	}

	Vector ret = new Vector();
	ret.addElement(new Integer(iret));
	ret.addElement(new Double(optmu));
	if ( (isRegularized) && (isRadoType) )
	    ret.addElement(new Double(optpena));

	return ret;
    }

    public double muBoost(int nf){
	// computes edge mu using operators
	int index;
	double sum = 0.0, vv, sig;
	double test_r, test_alpha;
	for (index=0;index<nRows;index++){
	    vv = (weights[index] * getOperator(isRadoType, split_number, index, nf)); //getOperator(index)[nf]);
	    sum += vv;
	}
	sum /= star[nf];

	if (Math.abs(sum) > 1.0)
	    Dataset.warning("|Mu| = " + Math.abs(sum) + " > 1.0 --- clamping to " + LinearBoost.MAX_ABS_MU + " ");
	if (Math.abs(sum) > LinearBoost.MAX_ABS_MU){
	    sum = Math.signum(sum) * LinearBoost.MAX_ABS_MU;
	}

	// Location of regularization for AdaBoost+L1, adapted from Xi, Xiang, Ramadge, Schapire (AISTATS'09)
	if ( (isRegularized) && (!isRadoType) ){
	    test_r = example_regularized_penalty_L1(sum, nf);
	    test_alpha = alphaBoost(muClamp(sum), nf);
	    if (Math.abs(test_alpha) > test_r){
		sum = invert_alphaBoost(Math.signum(test_alpha), test_r, nf);
	    }
	}

	return sum;
    }

    public double muClamp(double mu){
	if ( (methodMu == 1) && (Math.abs(mu) < LinearBoost.SCALE_MU_THRESHOLD) )
	    return (Math.signum(mu) * LinearBoost.SCALE_MU_THRESHOLD);
	else
	    return mu;
    }

    public boolean one_iteration(){
	// return true iff weights change, i.e. significant difference with update
	boolean sig;

	Vector nf = wfi();
	double mmu = ((Double) nf.elementAt(1)).doubleValue();
	int nnf = ((Integer) nf.elementAt(0)).intValue();

	double mu = muClamp(mmu);
	double alpha = alphaBoost(mu, nnf);

	theta[nnf] += alpha;
	sig = reweightAll(mu, nnf, alpha);
	return sig;
    }

    public double invert_alphaBoost(double ss, double alpha, int nf){
	double val = 0.0;
	double xx = 2.0 * star[nf] * alpha;
	double num = Math.exp(2.0 * star[nf] * alpha) - 1.0, den = Math.exp(2.0 * star[nf] * alpha) + 1.0;
	val = ss * num / den;

	return val;
    }

    public double alphaBoost(double mu, int nf){
	double val = 0.0;
	double num = 1.0 + mu, den = 1.0 - mu;
	val = Math.log(num/den);
	val /= (2.0 * star[nf]);

	return val;
    }

    public boolean reweightAll(double mu, int nf, double alp){
	if (methodWeights == 1)
	    return reweightAll_AdaBoostSS(mu, nf, alp);
	else if (methodWeights == 0)
	    return reweightAll_AdaBoostR(mu, nf, alp);
	else
	    Dataset.perror("No weighting method for " + methodWeights);
	return false;
    }

    public void normalizeWeights(){
	int index;
	double w, sum = 0.0, diff;

	for (index=0;index<nRows;index++){
	    w = weights[index];
	    sum += w;
	}

	if (sum < EPS)
	    Dataset.perror("sum of weights is approximately zero (" + sum + ")");

	for (index=0;index<nRows;index++){
	    w = weights[index];
	    w /= sum;
	    weights[index] = w;
	}

	// double check
	sum = 0.0;
	for (index=0;index<nRows;index++)
	    sum += weights[index];
	
	diff = Math.abs(1.0 - sum);
	if (diff > EPS)
	    Dataset.perror("LinearBoost :: sum of weights " + sum + " != 1.0");
    }

    public boolean reweightAll_AdaBoostSS(double mu, int nf, double alp){
	int index, i;
	double op, w, sum = 0.0, diff;
	boolean improvement;

	for (index=0;index<nRows;index++){
	    w = weights[index];
	    op = getOperator(isRadoType, split_number, index, nf); 
	    w = (w * Math.exp(- alp * op));
	    weights[index] = w;
	}
	normalizeWeights();
	improvement = (Math.abs(alp) > EPS2);

	return improvement;
    }

    public boolean reweightAll_AdaBoostR(double mu, int nf, double alp){
	double lastw, neww, den, num, op, diffc, maxd = 0.0, w;
	int index, i;
	boolean improvement;

	den = 1.0 - (mu * mu);

	for (index=0;index<nRows;index++){
	    lastw = weights[index];
	    w = lastw;
	    if (lastw <= 0.0)
		Dataset.perror("LinearBoost :: weights " + w + " < 0.0");
	    op = getOperator(isRadoType, split_number, index, nf); 
	    num = 1.0 - ( (mu * op) / star[nf]);
	    w = w * (num / den);
	    neww = w;

	    diffc = Math.abs(lastw - neww);
	    if ( (index == 0) || (diffc > maxd) )
		maxd = diffc;
	    
	    weights[index] = w;
	}

	normalizeWeights();
	// in theory, useless,
	// but kept for numerical imprecisions that otherwise may bring r_t out of bounds

	improvement = (maxd > 0.0);

	// stops boosting if no improvement
	if (Math.abs(alp) < EPS2)
	    improvement = false;
	if (oneWeight())
	  improvement = false;
	
	return improvement;
    }

    public boolean oneWeight(){
	// returns true iff all weight concentrated on one example, indicating a possibly trivial feature
	double w, maxw = 0.0;
	int index;

	for (index=0;index<nRows;index++){
	    w = weights[index];
	    if ( (index == 0) || (maxw < w) )
		maxw = w;
	}
	if (maxw > 1.0 - EPS){
	    return true;
	}
	return false;
    }

    // BOOSTING MISC FUNCTIONS

    public double getOperator(boolean isRado, int split, int indexRow, int indexCol){
	double v;
	if (isRado){
	    v = ((Rado) myDomain.rados.elementAt(indexRow)).operator[indexCol];
	    return v;
	}
	v = myDomain.myDS.train_example(split, indexRow).operator[indexCol];
	return v;
    }

    //Computes the error

    public double error(boolean onTraining){
	Example ee;
	double sumerr = 0.0;
	boolean rc;
	int i, ne;
	if (onTraining)
	    ne = myDomain.myDS.train_size(split_number);
	else
	    ne = myDomain.myDS.test_size(split_number);

	for (i=0;i<ne;i++){
	    if (onTraining)
		ee = myDomain.myDS.train_example(split_number, i);
	    else
		ee = myDomain.myDS.test_example(split_number, i);
	    rc = ee.rightClass(theta);
	    if (!rc)
		sumerr += 1.0;
	}
	sumerr /= (double) ne;

	return sumerr;
    }
}


/**************************************************************************************************************************************
 * Class Domain
 *****/

class Domain implements Debuggable{
    public static int MAX_RADO_SIZE = 1000;
    public static double Q_SLOPE = -1.0;
    public static double [] SLOPE_PHI;

    public MemoryMonitor myMemoryMonitor;

    Dataset myDS;
    Vector rados;

    int dimOperator, number_rados;

    public static String [] RADO_GENERATE = {"@RANDOM", "@N/A", "@RANDOM_PERCLASS"}; 

    public static int INDEX_GENERATE(String s){
	int i=0;
	do{
	    if (RADO_GENERATE[i].equals(s))
		return i;
	    i++;
	}while (i<RADO_GENERATE.length);
	Dataset.perror("Rado generation method " + s + " not found");
	return -1;
    }

    Domain(String nameD, String nameP, int maxr, double qs){ 
	SLOPE_PHI = null;
	double ar;
	int i;

	myMemoryMonitor = new MemoryMonitor();

	myDS = new Dataset(nameD, nameP, this);
	myDS.load_features();
	myDS.load_examples(this);
	dimOperator = ((Example) myDS.examples.elementAt(0)).operator.length;
	myDS.generate_stratified_sample_with_check(dimOperator);

	if (maxr > 0)
	    MAX_RADO_SIZE = maxr;

	if (qs >= 0.0){
	    // Slope regularisation: computation of the leveraging vector
	    // see for example http://statweb.stanford.edu/~candes/papers/SLOPE.pdf, or ArXiv http://arxiv.org/pdf/1512.05244v2.pdf

	    Q_SLOPE = qs;
	    SLOPE_PHI = new double [dimOperator];
	    for (i=0;i<dimOperator;i++){
		ar = 1.0 - ( ( (double) (i+1) ) * Q_SLOPE / (2.0 * (double) dimOperator ) );
		SLOPE_PHI[i] = Statistics.PhiInverse(ar); 
	    }
	}
    }

    public String memString(){
	return myMemoryMonitor.memString;
    }

    public String memStringThreshold(){
	return myMemoryMonitor.memStringThreshold;
    }

    public void rado_generate(int fold, LinearBoost lb){
	int nrad = lb.nRows;
	String method = lb.methodRados;
	double eps;

	int i, nt = 0, index = Domain.INDEX_GENERATE(method);
	boolean [] trivial;

	System.out.print("rados[" + nrad + "*" + Domain.RADO_GENERATE[index] + "");

	do{
	    if (index == 0){
		rado_generate_dumb(fold, nrad);
		nt = -1;
	    }else if (index == 1)
		Dataset.perror("Method " + method + " not in NIPS'16 paper. Please use another method to generate rados");
	    else if (index == 2){
		rado_generate_random_per_class(fold, nrad);
		nt = -1;
	    }
	}while(nt>0);
	System.out.print("ok.] ");
    }

    public void rado_generate_random_per_class(int fold, int nrad){
	Random r = new Random();
	rados = new Vector();
	int i, j, npos = 0, nneg = 0;
	double clref, minratio = -1.0, maxratio = -1.0, curratio;
	boolean pos;

	Vector indices;
	Rado dumr;

	for (i=0;i<nrad;i++){
	    indices = new Vector();
	    clref = Math.signum(myDS.train_example(fold,r.nextInt(myDS.train_size(fold))).normalized_class);
	    if (clref > 0){
		npos++;
		pos = true;
	    }else{
		nneg++;
		pos = false;
	    }

	    if ( (i == nrad -1) && (npos == 0) && (!pos) )
		pos = true;
	    else if ( (i == nrad -1) && (nneg == 0) && (pos) )
		pos = false;
		

	    for (j=0;j<myDS.train_size(pos,fold);j++){
		if (r.nextBoolean())
		    indices.addElement(new Integer(j));
	    }

	    dumr = new Rado(i, myDS, pos, fold, indices);

	    rados.addElement(dumr);
	    displayProcess(i, nrad);
	}
	//System.out.print("ok...");
    }

    public void displayProcess(int i, int nrad){
	if (nrad >= 10){
	    if (i%(nrad/10) == 0)
		System.out.print(".");
	    //System.out.print( ((i/(nrad/10)) * 10) + "% ");
	}else 
	    if (i%(nrad/2) == 0)
		System.out.print(".");
	//System.out.print( ((i/(nrad/2)) * 2) + "% ");
    }

    public void rado_generate_dumb(int fold, int nrad){

	Random r = new Random();
	rados = new Vector();

	Rado meano = new Rado(-1, myDS, fold);	
	int i, j;
	double [] rade = new double [myDS.train_size(fold)];
	double clref, minratio = -1.0, maxratio = -1.0, curratio;

	Rado dumr;

	for (i=0;i<nrad;i++){
	    for (j=0;j<myDS.train_size(fold);j++){
		if (r.nextBoolean())
		    rade[j] = 1.0;
		else
		    rade[j] = -1.0;
	    }

	    dumr = new Rado(i, myDS, rade, fold, meano);

	    rados.addElement(dumr);
	    displayProcess(i, nrad);
	}
    }

    public double radostar(LinearBoost lb, int nv){
	//returns the largest absolute magnitude in rado fold

	double val = 0.0, dcv, cv2;
	Rado ce;
	int i;
	for (i=0;i<rados.size();i++){
	    dcv = lb.getOperator(true, lb.split_number, i, nv); 

	    cv2 = Math.abs(dcv);
	    if ( (i==0) || (cv2 > val) )
		val = cv2;
	}
	return val;
    }


    public void save_operator_and_class_examples(){
	String saveFileName = myDS.pathSave + myDS.domainName + "_domain.operator.txt";
	FileWriter f = null;
	int i, j;
	double [] curop;

	try{
	    f = new FileWriter(saveFileName);
	    f.write("// Domain + " + myDS.domainName + ", operators and classes in whole domain\n");
	    f.write("// Total #examples : " + myDS.number_examples_total + ", #features excl. class : " + myDS.number_initial_features + "\n");
	    for (i=0;i<myDS.number_examples_total;i++){
		curop = ((Example) myDS.examples.elementAt(i)).operator;
		for (j=0;j<curop.length;j++){
		    f.write(curop[j] + "");
		    f.write(Dataset.KEY_SEPARATION_STRING[Dataset.SEPARATION_INDEX]);
		}
		f.write(((Example) myDS.examples.elementAt(i)).normalized_class + "\n");
	    }
	    f.close();
	}catch(IOException e){
	    Dataset.perror("LinearAdoboost.class :: Saving examples/domain error in file " + saveFileName);
	}
    }

    public void safeCheck(){
	//does nothing
    }
}
